import { deepValue } from "./utils/deepValue";
// @ts-ignore: This will be placed in dist on build.
import { search as wasmSearch } from "./wafu";
export const defaultOptions = {
    location: 0,
    distance: 100,
    threshold: 0.6,
    caseSensitive: false,
    keys: [],
    shouldSort: true,
    getFn: deepValue,
    includeMatches: false,
    minMatchCharLength: 1,
    tokenize: false,
    tokenSeparator: / +/g,
    matchAllTokens: false
};
export class Wafu {
    constructor(collection, opts) {
        this.options = Object.assign({}, defaultOptions, opts);
        this.collection = collection;
        this.fields = buildFields(this.collection, this.options);
        this.rsFields = buildRustFields(this.fields);
        this.rsOptions = buildRustOptions(this.options);
    }
    search(pattern, opts) {
        if (pattern === "" || this.fields.length === 0) {
            return [];
        }
        if (!this.options.caseSensitive) {
            pattern = pattern.toLowerCase();
        }
        // Here we're going to deviate from fuse a bit, only because fuse's
        // solution is hard for me to implement and this shouldn't come up too
        // often. When the pattern or any of the pattern tokens are longer than 32
        // chars, just truncate to 32 chars. This is a limitation of bitap (kind
        // of). Think more on this later, but avoids a runtime error for now.
        let pattern_tokens = null;
        if (this.options.tokenize) {
            pattern_tokens = pattern
                .split(this.options.tokenSeparator)
                .filter(v => v.length > 0)
                .map(s => truncateTo32Chars(s));
        }
        const input = {
            pattern: truncateTo32Chars(pattern),
            pattern_tokens,
            limit: opts && typeof opts.limit === "number" ? opts.limit : null,
            fields: this.rsFields,
            options: this.rsOptions
        };
        // @ts-ignore: Can't use a function that ts doesn't know about!
        const rsResults = wasmSearch(input);
        return buildJSResult(this.collection, this.fields, rsResults, this.options);
    }
}
function truncateTo32Chars(s) {
    const chars = Array.from(s);
    if (chars.length < 31)
        return s;
    const truncated = chars.slice(0, 30).join("");
    return truncated;
}
// Takes the original collection and the passed options and makes a
// normalized/flat version of each searchable field. This is an internal
// pre-processing step that makes passing to rust a little more simple.
function buildFields(collection, opts) {
    // Normalize keys into a {name, weight} structure.
    let keys = opts.keys.map(key => {
        if (typeof key === "string") {
            return { name: key, weight: 1 };
        }
        else {
            return { name: key.name, weight: key.weight };
        }
    });
    // Flatten the list of strings that we want to search down into a single
    // array, with references back to their original "locations" as determined
    // by keyIndex and arrayIndex.
    let fields = [];
    collection.forEach((item, itemIndex) => {
        // Handle the simple string case.
        if (typeof item === "string") {
            fields.push({
                originalText: item,
                text: item,
                itemIndex,
                weight: 1
            });
            return;
        }
        // Handle the object case.
        keys.forEach((key, keyIndex) => {
            const value = opts.getFn(item, key.name);
            if (typeof value === "string") {
                fields.push({
                    originalText: value,
                    text: value,
                    itemIndex,
                    keyIndex,
                    key: key.name,
                    weight: key.weight
                });
            }
            else if (Array.isArray(value)) {
                // Technically fuse allows recursive array lookups, but my getFn
                // definition is a little more strict so I'm going to ignore.
                value.forEach((subvalue, arrayIndex) => {
                    if (typeof subvalue !== "string") {
                        return;
                    }
                    fields.push({
                        originalText: subvalue,
                        text: subvalue,
                        itemIndex,
                        keyIndex,
                        arrayIndex,
                        key: key.name,
                        weight: key.weight
                    });
                });
            }
        });
    });
    // Filter out empty fields. I think this is ok, since these fields would
    // almost certainly not be matches, but not _really_ sure.
    fields = fields.filter(field => field.text !== "");
    // Convert all text to lowercase when case sensitive.
    if (!opts.caseSensitive) {
        for (let field of fields) {
            field.text = field.text.toLowerCase();
        }
    }
    // Tokenize all strings when tokenize is true.
    if (opts.tokenize) {
        for (let field of fields) {
            // TODO: Could skip if the result of the split is a single string that's
            // equal to the original string, ie: "fuse".split(/ +/g) is just
            // ["fuse"]. The scores for full and token are averaged, so ultimately
            // the score will be the same if we just don't run on tokens.
            //
            // However, after dealing with the analyze function more deeply I think
            // any optimization like this would change the resulting scores pretty
            // drastically. Will need to think more on this later.
            // NOTE: Leading and trailing whitespace in the text may cause
            // extraneous empty tokens. We filter these out. Ex:
            //
            //   " fuse ".split(/ +/g) returns ["", "fuse", ""]
            //
            // Fuse notably doesn't handle this, but I'm desperately hoping no one
            // is depending on that behavior.
            field.tokens = field.text
                .split(opts.tokenSeparator)
                .filter(v => v.length > 0);
        }
    }
    return fields;
}
function buildRustOptions(opts) {
    return {
        location: opts.location,
        distance: opts.distance,
        threshold: opts.threshold,
        include_matches: opts.includeMatches,
        min_match_char_length: opts.minMatchCharLength,
        tokenize: opts.tokenize,
        match_all_tokens: opts.matchAllTokens,
        should_sort: opts.shouldSort
    };
}
function buildRustFields(fields) {
    return fields.map(f => ({
        text: f.text,
        tokens: f.tokens ? f.tokens : null,
        item_index: f.itemIndex,
        weight: f.weight
    }));
}
function buildJSResult(collection, fields, results, opts) {
    return results.map(rsResult => {
        const jsResult = {
            item: collection[rsResult.item_index],
            score: rsResult.score
        };
        if (opts.includeMatches && rsResult.matches !== null) {
            jsResult.matches = rsResult.matches.map(rsMatch => {
                const field = fields[rsMatch.field_index];
                const jsMatch = {
                    indices: rsMatch.indices,
                    value: field.originalText
                };
                if (field.key !== undefined) {
                    jsMatch.key = field.key;
                }
                if (field.arrayIndex !== undefined) {
                    jsMatch.arrayIndex = field.arrayIndex;
                }
                return jsMatch;
            });
        }
        return jsResult;
    });
}
